// -*- mode: c++; coding: utf-8 -*-

// Linthesia

// Copyright (c) 2007 Nicholas Piegdon
// Adaptation to GNU/Linux by Oscar Aceña
// See COPYING for license information

#ifndef __OS_GRAPHICS_H
#define __OS_GRAPHICS_H

// only for 'Ms. pepis' debugging
#include <iostream>

#include <gtkmm.h>
#include <gtkglmm.h>

#include <GL/gl.h>
#include <GL/glu.h>

#endif // __OS_GRAPHICS_H
