// -*- mode: c++; coding: utf-8 -*-

// Linthesia

// Copyright (c) 2007 Nicholas Piegdon
// Adaptation to GNU/Linux by Oscar Aceña
// See COPYING for license information

#ifndef __LINTHESIA_VERSION_H__
#define __LINTHESIA_VERSION_H__

#include <string>

// See CHANGELOG for a list of what has changed between versions
static const std::string LinthesiaVersionString = "0.4.3";

#endif // __LINTHESIA_VERSION_H__
