// -*- mode: c++; coding: utf-8 -*-

// Linthesia

// Copyright (c) 2007 Nicholas Piegdon
// Adaptation to GNU/Linux by Oscar Aceña
// See COPYING for license information

#ifndef __TEXTURES_H
#define __TEXTURES_H

enum Texture {

  TitleLogo,
  InterfaceButtons,

  ButtonRetrySong,
  ButtonChooseTracks,
  ButtonExit,
  ButtonBackToTitle,
  ButtonPlaySong,

  InputBox,
  OutputBox,
  SongBox,

  TrackPanel,

  StatsText,

  PlayStatus,
  PlayStatus2,
  PlayKeys,

  PlayNotesBlackColor,
  PlayNotesBlackShadow,
  PlayNotesWhiteColor,
  PlayNotesWhiteShadow,

  PlayKeyRail,
  PlayKeyShadow,
  PlayKeysBlack,
  PlayKeyHits,

  _TextureEnumCount
};

#endif // __TEXTURES_H
